/*
 * Copyright (c) 2008, Noel Lemouel
 * All rights reserved. See LICENCE file for details.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#define EB 0x00
#define EL 0x01

static int host_endianness;
static int target_endianness;

static int sc_open(FILE**, const char*, struct stat*);
static int so_create(FILE**, const char*);
static void usage(char*);

void endianness_detect(void)
{
   unsigned int ie = 0x00000001;

   if(*(unsigned char*)&ie){
      printf("[little endian host]\n");
      host_endianness = EL;
   }else{
      printf("[big endian host]\n");
      host_endianness = EB;
   }
}

unsigned int swap32(unsigned int xe)
{
	int i=4;
	unsigned int ye = 0;
   
	while(i--){
		ye <<= 8;
		ye |= xe & 0xff;
		xe >>= 8;
	}

	return ye;
}

unsigned int el_swap32(unsigned int le)
{
	int i=4;
	unsigned int be = 0;

   if(host_endianness) return le;
   
	while(i--){
		be <<= 8;
		be |= le & 0xff;
		le >>= 8;
	}

	return be;
}

unsigned int el_swap16(unsigned int le)
{
	int i=2;
	unsigned int be = 0;

   if(host_endianness) return le;
   
	while(i--){
		be <<= 8;
		be |= le & 0xff;
		le >>= 8;
	}

	return be;
}

struct bmp_file_hdr{
	unsigned char magic[2];	/* BM */
	unsigned int size;
	unsigned short reserved_0;
	unsigned short reserved_1;
	unsigned int data_offset;
} __attribute__((__packed__));

struct dib_hdr{
	unsigned int hdr_size;	/* expects 40bytes exits otherwise */
	int bmp_width;
	int bmp_height;
	unsigned short n_color_planes;
	unsigned short n_bpp;	/* expects 24 */
	unsigned int compression;	/* expects 0 (BI_RGB) exits otherwise */
	unsigned int bmp_size;
	int h_res;	/* pixel per meter */
	int v_res;	/* pixel per meter */
	unsigned int n_palette_colors;	/* ignored */
	unsigned int n_valid_colors;	/* ignored */
} __attribute__((__packed__));

struct rawx_hdr{
	unsigned char magic[4];
	unsigned int width;
	unsigned int height;
	unsigned int mode;	/* ARGB32 */
	unsigned int reserved;
} __attribute__((__packed__));

static struct bmp_file_hdr* bmp_file_info;
static struct dib_hdr* dib_info;

int rawx_write_header(FILE** fd)
{
	unsigned char *c;
	struct rawx_hdr rawx_info;

	if(target_endianness == EL)
	{
		strncpy((char*)&(rawx_info.magic), "RAWL", 4);
		rawx_info.width = dib_info->bmp_width;
		rawx_info.height = dib_info->bmp_height;
		rawx_info.mode = host_endianness ? 32 : swap32(32); /* temporary */
		rawx_info.reserved = 0x00;
	}else{	/* EB */
		strncpy((char*)&(rawx_info.magic), "RAWB", 4);
		rawx_info.width = swap32(dib_info->bmp_width);
		rawx_info.height = swap32(dib_info->bmp_height);
		rawx_info.mode = host_endianness ? swap32(32) : 32; /* temporary */
		rawx_info.reserved = 0x00;
	}

	/* debug */
   c = (unsigned char *)&rawx_info;
	printf("__rawx header__\n");
	printf("magic   : %c%c%c%c\n", *(c), *(c+1), *(c+2), *(c+3)); c+=4;
	printf("width   : 0x%02x%02x%02x%02x\n", *(c), *(c+1), *(c+2), *(c+3)); c+=4;
	printf("height  : 0x%02x%02x%02x%02x\n", *(c), *(c+1), *(c+2), *(c+3)); c+=4;
	printf("mode    : 0x%02x%02x%02x%02x\n", *(c), *(c+1), *(c+2), *(c+3)); c+=4;
	printf("reserved: 0x%02x%02x%02x%02x\n", *(c), *(c+1), *(c+2), *(c+3)); c+=4;

	/* write header to the file */
	fwrite((const void*)&rawx_info, sizeof(struct rawx_hdr), 1, *fd);
}

int rawx_write_line_argb32(FILE** fd, unsigned char* buf, unsigned int len)
{
	unsigned int i = len >> 2;
	unsigned int* buf32 = (unsigned int*)buf;

	/* buf contains le data (bgra) */
	if(target_endianness == EL){
		fwrite((const void*)buf, len, 1, *fd);
	}else{	/* EB */
		while(i--)
		{
			*buf32 = swap32(*buf32);
			buf32++;
		}
		fwrite((const void*)buf, len, 1, *fd);
	}
}

int bmp2rawx(FILE** fd, unsigned char* bmp)
{
	unsigned int i, j, padding = 0;
	unsigned char *line, *pxl;
	unsigned char *l_buf;

	bmp_file_info = (struct bmp_file_hdr*)bmp;
	dib_info = (struct dib_hdr*)((int)bmp + (int) sizeof(struct bmp_file_hdr));

	if(strncmp((char*)bmp_file_info->magic, "BM", 2)){
		printf("error: invalid magic\n");
		return -1;
	}

	printf("size: %d\n", el_swap32(bmp_file_info->size));

	if(el_swap32(dib_info->hdr_size) != 40)
	{
		printf("error: unsuported format [dib_info->hdr_size]\n");
		return -1;
	}

	printf("geometry: %dx%dx%d\n", \
				el_swap32(dib_info->bmp_width),
				el_swap32(dib_info->bmp_height),
				el_swap16(dib_info->n_bpp));

	if(el_swap16(dib_info->n_bpp) != 24)
	{
		printf("error: 24bits file format only [read: %d]\n", (dib_info->n_bpp)>>8);
		return -1;
	}

	/* estimate padding */

	if((el_swap32(dib_info->bmp_width) * 3) % 4){
		padding = ((((el_swap32(dib_info->bmp_width) * 3) >> 2) + 1) << 2) - (el_swap32(dib_info->bmp_width) * 3);
		printf("padding: %d/row\n", padding);
	}

	/* allocate a 1 line length buffer (argb32 / 32bpp) */

	l_buf = malloc(el_swap32(dib_info->bmp_width) << 2);
	if(!l_buf){
		printf("error: [alloc. %d] failed\n", el_swap32(dib_info->bmp_width) << 2);
		return -1;
	}else{
		printf("[alloc. %d]\n", el_swap32(dib_info->bmp_width) << 2);
	}

   /* point to the bitmap data array and go to the first line (last in file) */
 
	i = el_swap32(dib_info->bmp_height);
   
	line = (unsigned char*)((int)bmp  \
		   + (int)el_swap32(bmp_file_info->data_offset)	\
		   + ((el_swap32(dib_info->bmp_height) - 1)	\
             * ((el_swap32(dib_info->bmp_width) * 3) + padding)) );

	rawx_write_header(fd);

	while(i--)
	{
		pxl = line;
		j = el_swap32(dib_info->bmp_width);    

		while(j--)
		{
			/* keep native bmp data byte order (le) */
			/* add alpha channel */
			l_buf[0] = pxl[0];   /* blue */
			l_buf[1] = pxl[1];   /* green */
			l_buf[2] = pxl[2];   /* red */
			l_buf[3] = 0xff;     /* alpha */
			l_buf += 4;
			pxl += 3;
		}
		l_buf -= (el_swap32(dib_info->bmp_width) << 2);
		rawx_write_line_argb32(fd, l_buf, el_swap32(dib_info->bmp_width) << 2);
		line -= (el_swap32(dib_info->bmp_width) * 3) + padding;
	}

	free(l_buf);
  
	return 0;
}

int main(int argc, char* argv[])
{
	FILE* fd0;
	FILE* fd1;
	struct stat info;

	unsigned char* bmp_buf;
	unsigned long address;

	if(argc < 3){
		usage(argv[0]);
		return -1;
	}else{
		if(!strcmp(argv[1], "-le")){
			target_endianness = EL;
      		printf("[little endian target]\n");
		}else if(!strcmp(argv[1], "-be")){
			target_endianness = EB;
      		printf("[big endian target]\n");
		}else{
			usage(argv[0]);
			return -1;
		}
		if( sc_open(&fd0, argv[2], &info) ) return -1;
		if( so_create(&fd1, argv[3]) ) return -1;
	}

	/* min 54 bytes (file header + DIB header (Windows V3 	BITMAPINFOHEADER)) */

	if(info.st_size < 54){
		printf("error: file is too small [%d]\n", info.st_size);
		return -1;
	}

	bmp_buf = malloc(info.st_size);
	if(!bmp_buf){
		printf("error: [alloc. %d] failed\n", (int)info.st_size);
		return -1;
	}else{
		printf("[alloc. %d]\n", (int)info.st_size);
	}

	/* ... */

	endianness_detect();
	fread(bmp_buf, 1, info.st_size, fd0);
	bmp2rawx(&fd1, bmp_buf);

	/* ... */

	free(bmp_buf);
	fclose(fd0);
	fclose(fd1);

	return 0;
}

static int sc_open(FILE** fd, const char* name, struct stat* info)
{
      if (stat(name, info) == -1)
      {
          if (errno == ENOENT)
          {
            (void) fprintf( stderr, "error: '%s' does not exist\r\n", name );
          }else{
            (void) fprintf( stderr, "error: stat() fails for '%s'\r\n", name );
          }
          return -1;
      }

      /* checking S_ISLNK(m) is not required since 'stat' follows sym. links */
      if (!S_ISREG(info->st_mode))
      {
         (void) fprintf( stderr, "error: '%s' is a not a regular file (or a sym. link)\r\n", name);
         return -1;
      }

      if ((*fd = fopen(name, "r")) == NULL)
      {
         (void) fprintf( stderr, "error: '%s' open failed\r\n", name );
         return -1;
      }

	  /* printf("size: %d\n", (int)info->st_size); */

      return 0;   /* no error */
}

static int so_create(FILE** fd, const char* name)
{
	struct stat so;
	unsigned char c;

    if (stat(name, &so) == -1)
    {
		if (errno == ENOENT)
        {
            /* does not exist => create it */
			if ((*fd = fopen(name, "w+")) == NULL)
      		{
         		(void) fprintf( stderr, "error: '%s' open failed\r\n", name );
         		return -1;
			}
        }else{
            (void) fprintf( stderr, "error: stat() fails for '%s'\r\n", name );
        	return -1;
        }
    }else if (S_ISREG(so.st_mode)){
		/* file exist => overwrite ? */		
		while(c != 'y' && c != 'n'){
			printf("%s is already existing, overwrite (y/n) ? ", name);

			c = 'y'; /* getchar(); fix this */
			printf("\n");

			/*
				#include <termios.h>
				...
				struct termios t0, t1;
				tcgetattr(0, &t0);
				...
				t1.c_lflag &= ~ECHO; // echo off
				t1.c_lflag &= ~ICANON; // one char at a time
			*/
		}
		if(c == 'y')
		{
      		if ((*fd = fopen(name, "w+")) == NULL)
      		{
         		(void) fprintf( stderr, "error: '%s' open failed\r\n", name );
         		return -1;
			}
		}else{
			return -1;
		}
	}

    return 0;   /* no error */
}

static void usage(char* s)
{
	printf("usage: %s -le | -be input.bmp output.rawx\n", s);
}
